#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <io.h>

#include "modsetup.h"

#define GAMESCOUNT 2
#define INFOSCOUNT 7
#define FIXESCOUNT 27

#pragma pack(push, 1)
typedef struct {
  DWORD offs[GAMESCOUNT];
  BYTE  data[2];
} fix_item;

typedef struct {
  char *name;
  char *game;
  DWORD size;
} fix_list;

typedef struct {
  BYTE idx;
  BYTE count;
  BYTE stat;
} fix_info;

typedef struct {
  char  name[12];
  DWORD size;
} grp_item;
#pragma pack(pop)

fix_list gamelist[GAMESCOUNT] = {
  {"EX.EXE", "Exhumed",    825341},
  {"PS.EXE", "PowerSlave", 818966}
};

fix_info infolist[INFOSCOUNT] = {
  /* Revert [Run] logic (always run) */
  { 0,  1, 0},
  /* Always [Mouseview] */
  { 1,  4, 0},
  /* Revert [Strafe] logic */
  { 5,  4, 0},
  /* /CDOFF always enabled (nocd) */
  { 9,  2, 0},
  /* Allow F12 key to take screenshots */
  {11,  8, 0},
  /* Fix C key to work properly */
  {19,  1, 0},
  /* Custom BUILD screen resolutions */
  {20,  7, 0}
};

/* screen resolution hack item index */
#define SCR_IDX 6

fix_item datalist[FIXESCOUNT] = {
  /* Run (7) - revert logic */
  {0x000367FE, 0x0003639B, 0x74, 0x75}, /* always run */
  /* Mouseview (25) - always on */
  {0x00036887, 0x00036424, 0xE9, 0xB8}, /* [3] allow mouseview when Move_Forward */
  {0x000368B1, 0x0003644E, 0xE9, 0xBA}, /* [4] allow mouseview when Move_Backward */
  {0x000368EF, 0x0003648C, 0x42, 0x00}, /* [1] always mouseview */
  {0x00039A7E, 0x000395B9, 0x02, 0x00}, /* [2] turn off center view */
  /* Strafe (4) - revert logic (always strafe; faster than strafe left / strafe right) */
  {0x00036A81, 0x0003661E, 0x84, 0x85}, /* [1] revert logic */
  {0x00036AA6, 0x00036643, 0x87, 0x57}, /* [2] go to mouse_look_turn */
  {0x00036AB0, 0x0003664D, 0x21, 0x1C}, /* [4] go to [3] */
  {0x00036ACE, 0x0003666B, 0x5F, 0x2F}, /* [3] go to mouse_look_turn */
  /* /cdoff always enabled */
  {0x000A2EC2, 0x000A1B8A, 0x00, 0x01}, /* nocd */
  /* that patch byte needed only for Exhumed - skips MSCDEX check */
  {0x0005AB77, 0x00000000, 0x75, 0xEB},
  /* v1.2 - screenshots and PC Speaker patch */
  /* disable PC Speaker (A LOT faster screenshot save) */
  {0x00066092, 0x00065B84, 0xBA, 0xEB},
  {0x00066093, 0x00065B85, 0x61, 0x07},
  {0x000661EC, 0x00065CDE, 0xBA, 0xEB},
  {0x000661ED, 0x00065CDF, 0x61, 0x0A},
  /* EX: cmp b,[00273AC9], 000 -> test b,[00273EB8], 000 */
  /* PS: cmp b,[00273AF5], 000 -> test b,[00273EE4], 000 */
  /* relocation table */
  {0x000046EC, 0x80004762, 0xC9, 0xB8}, /* PS: F5 -> E4 */
  {0x000047BB, 0x80004772, 0xC9, 0xB8},
  {0x0002FC36, 0x8002F792, 0xC9, 0xB8},
  {0x0002FC4D, 0x8002F7A9, 0xC9, 0xB8},
  /* v1.2 - fix C key to work properly */
  {0x0002DFF6, 0x0002DBF7, 0x27, 0x34},
  /* v1.1 - vid_hack */
  /* always hi-res mode as "ScreenWidth = 640" in SETUP.CFG */
  {0x00049D64, 0x000499B7, 0x05, 0x00},
  /* read "STATUS.MSK" from current directory instead of STUFF.DAT */
  {0x0005DC46, 0x0005D73B, 0x01, 0x00},
  /* disable x2 scale for hires */
  {0x0005DCFC, 0x0005D7F1, 0x74, 0xEB},
  /* width */
  {0x0002EA6C, 0x0002E5EC, 0x80, 0x80},
  {0x0002EA6D, 0x0002E5ED, 0x02, 0x02},
  /* height */
  {0x0002EA71, 0x0002E5F1, 0x90, 0x90},
  {0x0002EA72, 0x0002E5F2, 0x01, 0x01}
};

char strData[] = "STUFF.DAT";
char strMask[] = "STATUS.MSK";
char strSign[] = "KenSilverman";
/* max width: 1600 */
WORD maskdata[1600];
WORD masksize;

menuitem maingame[GAMESCOUNT + 1];

menuitem mainmenu[] = {
  {
    /* menu title */
    "Main Menu",
    /* subheader (green) */
    NULL,
    /* total menu items (this not included) */
    5
  },
  {
    /* menu item name */
    "Hacks Setup",
    /* menu item hint */
    "Configure various hacks.",
    /* menu item flag */
    0
  },
  {
    "Custom resolution",
    "Allow to set custom BUILD screen resolution.",
    0
  },
  {
    "About DeHacker",
    "Information about this program.",
    0
  },
  {
    "Save settings and exit",
    "Save the current settings and exit.",
    0
  },
  {
    "Exit without saving settings",
    "Exit without saving any changes.",
    0
  }
};

menuitem mainctrl[] = {
  {"Hacks Menu", NULL, 6},
  {
    "Revert [Run] logic (always run)",
    "Always run when [Run] key NOT pressed.",
    ITEM_CHKBOX
  },
  {
    "Always [Mouseview]",
    "Mouse look will be always enabled. This also fixes look and walk behaviour.",
    ITEM_CHKBOX
  },
  {
    "Revert [Strafe] logic (always strafe)",
    "Strafe always enabled when [Strafe] key NOT pressed. Allows fast strafe.",
    ITEM_CHKBOX
  },
  {
    "/CDOFF always enabled (nocd)",
    "The game will work without CD. This also disables MSCDEX check for Exhumed.",
    ITEM_CHKBOX
  },
  {
    "Allow F12 key to take screenshots",
    "Enables native screenshot support and disables PC Speaker noises (FAST save).",
    ITEM_CHKBOX
  },
  {
    "Fix C key to work properly",
    "Stabilizes C key so it can be used for playing (Alt+C implementation bug).",
    ITEM_CHKBOX
  }
};

char strBroken[] = "Broken status bar (HUD) due to status bar mask buffer limitations.";
menuitem mainmode[] = {
  {"Screen Resolution Menu", NULL, 14},
  {"Disabled",  "Disables hi-res hack and revert all changes back.", ITEM_HILITE | ITEM_RADBOX},
  { "320x200",  NULL, ITEM_RADBOX},
  { "360x200",  NULL, ITEM_RADBOX},
  { "320x240",  NULL, ITEM_RADBOX},
  { "360x240",  NULL, ITEM_RADBOX},
  { "320x400",  NULL, ITEM_RADBOX},
  { "360x400",  NULL, ITEM_RADBOX},
  { "640x350",  NULL, ITEM_RADBOX},
  { "640x400",  NULL, ITEM_RADBOX},
  { "640x480",  NULL, ITEM_RADBOX},
  { "800x600",  NULL, ITEM_RADBOX},
  {"1024x768",  NULL, ITEM_RADBOX},
  {"1280x1024", strBroken, ITEM_RADBOX},
  {"1600x1200", strBroken, ITEM_RADBOX}
};

menuitem maininfo[] = {
  {"About DeHacker", NULL, 12 | MENU_NFOBOX},
  {"Exhumed / PowerSlave DeHacker Version 1.2", NULL, 0},
  {"(c) -=CHE@TER=- 2014,2017", NULL, 0},
  {"e-mail: _CTPAX_@MAIL.RU", NULL, 0},
  {"http://ctpax-cheater.losthost.org/htmldocs/ps_vs_ex.htm", NULL, 0},
  {"", NULL, 0},
  {"Special thanks to:", NULL, 0},
  {"", NULL, 0},
  {"Corvin", NULL, 0},
  {"http://www.r-t-c-m.com", NULL, 0},
  {"", NULL, 0},
  {"BlackDoomer", NULL, 0},
  {"https://cher-nov.github.io", NULL, 0}
};

menuitem mainexit[] = {
  {"Save settings?", NULL, 2 | MENU_DLGBOX},
  {"Yes", "Save current settings and exit program.", 0},
  {"No", "Do not save current settings, but exit program.", 0}
};

int MenuProc(int item) {
int r;
  r = 1;
  switch (item) {
    /* hacks */
    case 1:
      DrawMenu(mainctrl, NULL);
      break;
    /* video resolutions */
    case 2:
      DrawMenu(mainmode, NULL);
      break;
    /* about */
    case 3:
      DrawMenu(maininfo, NULL);
      break;
    /* exit */
    case 4:
    case 5:
      r = 0;
      break;
  }
  return(r);
}

/*
  FILE and fopen(), fclose(), etc.
  adds a lot of code
  so avoid to use that function
  use open() instead
*/

int FileExists(char *fname, DWORD fsize) {
int fl;
  fl = open(fname, O_RDONLY | O_BINARY);
  if (fl != -1) {
    if (fsize) {
      lseek(fl, 0, SEEK_END);
      fsize = (tell(fl) == fsize) ? 1 : 0;
    } else {
      fsize = 1;
    }
    close(fl);
  } else {
    fsize = 0;
  }
  return(fsize);
}

/*
  atoi() adds a lot of code
  (mostly the large character table for the isdigit() function)
  so avoid to use that function
*/
WORD StrToInt(char *s) {
WORD r;
  r = 0;
  if (s) {
    for (; *s; s++) {
      if ((*s >= '0') && (*s <= '9')) {
        r *= 10;
        r += (*s - '0');
      } else {
        break;
      }
    }
  }
  return(r);
}

/*
  compare two strings
  c - NOT null terminated string, max 12 characters
  s - ASCIIZ null terminated string
*/
int MatchItem(char *c, char *s) {
int i, r;
  r = 1;
  for (i = 0; i < 12; i++) {
    /* not equal chars - strings not matched */
    if (*c != *s) {
      r = 0;
      break;
    }
    /* string ends earlier */
    if (!*c) { break; }
    /* go to next chars */
    c++;
    s++;
  }
  /* if 12 chars exceeded
     and there are still
     chars left in s */
  if (*s) {
    r = 0;
  }
  return(r);
}

/* actually "STATUS.MSK" can be here as static array but since anyone
   can mod it - take the actual version from the "STUFF.DAT" file */
void ReadBuildMask(WORD width, int state) {
grp_item item;
DWORD i, sz, fsz;
int fl;
  switch (state) {
    /* read mask */
    case 0:
      masksize = 0;
      fl = open(strData, O_RDONLY | O_BINARY);
      if (fl != -1) {
        lseek(fl, 0, SEEK_END);
        fsz = tell(fl);
        lseek(fl, 0, SEEK_SET);
        /* sanity check */
        if (fsz >= sizeof(item)) {
          read(fl, &item, sizeof(item));
          /* header size */
          sz = (item.size + 1) * sizeof(item);
          /* some more sanity checks */
          if (MatchItem(item.name, strSign) && (sz <= fsz)) {
            for (i = item.size; i; i--) {
              read(fl, &item, sizeof(item));
              /* found * few more sanity checks */
              if (MatchItem(item.name, strMask) && (item.size >= 4) && ((sz + item.size) <= fsz)) {
                /* go to file */
                lseek(fl, sz, SEEK_SET);
                /* mask size */
                read(fl, &masksize, 2);
                /* unused */
                read(fl, &fsz, 2);
                /* max size */
                sz = (item.size - 4);
                if (masksize > sz) {
                  masksize = sz;
                }
                /* size in bytes => elements count */
                masksize /= 2;
                if (masksize > 1600) {
                  masksize = 1600;
                }
                /* read mask */
                read(fl, &maskdata[0], masksize * 2);
                break;
              }
              /* calc absolute item offset */
              sz += item.size;
            }
          }
        }
        close(fl);
      }
      break;
    /* build mask */
    case 1:
      if (masksize) {
        fl = open(strMask, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
        if (fl != -1) {
          /* mask can't be used for resolutions bigger than 1024 - game will crash */
          if (width > 1024) {
            /* at least hide mask so it won't be drawn as ugly blocks at left */
            width = 320;
            for (i = 0; i < width; i++) {
              maskdata[i] = 200;
            }
          }
          i = width*2;
          /* size */
          write(fl, &i, 2);
          /* unused */
          i = 0;
          write(fl, &i, 2);
          /* multiply by 1000 to avoid float coeffs */
          sz = width;
          sz = (sz * 1000) / masksize; /* (width * 1000) / 320 */
          /* extend mask */
          for (i = 0; i < width; i++) {
            write(fl, &maskdata[(i * 1000) / sz], 2);
          }
          close(fl);
        }
      }
      break;
    /* delete mask */
    case 2:
      if (FileExists(strMask, 0)) {
        remove(strMask);
      }
      break;
  }
}

void LoadPatches(int n) {
WORD w, h, x, y;
int fl, i;
char *s;
BYTE b;
  /* load mask */
  ReadBuildMask(0, 0);
  fl = open(gamelist[n].name, O_RDONLY | O_BINARY);
  if (fl != -1) {
    for (i = 0; i < INFOSCOUNT; i++) {
      /* lazy check - only first byte */
      lseek(fl, datalist[infolist[i].idx].offs[n], SEEK_SET);
      read(fl, &b, 1);
      infolist[i].stat = (b == datalist[infolist[i].idx].data[0]) ? 0 : 1;
      /* menu selection */
      if (i < mainctrl[0].flag) {
        mainctrl[i + 1].flag |= infolist[i].stat ? ITEM_HILITE : 0;
      }
    }
    /* mask not loaded */
    if (!masksize) {
      infolist[SCR_IDX].stat = 0;
      mainmode[0].flag = 1 | MENU_NFOBOX;
      mainmode[0].item = "Custom resolution hack disabled";
      mainmode[0].hint = NULL;
      mainmode[1].flag = ITEM_HILITE;
      mainmode[1].item = "Status bar mask file not found.";
      mainmode[1].hint = NULL;
    }
    /* screen resolutions hack */
    if (infolist[SCR_IDX].stat) {
      lseek(fl, datalist[infolist[SCR_IDX].idx + 3].offs[n], SEEK_SET);
      read(fl, &w, 2);
      lseek(fl, datalist[infolist[SCR_IDX].idx + 5].offs[n], SEEK_SET);
      read(fl, &h, 2);
      infolist[SCR_IDX].stat = 0;
      for (i = 1; i < mainmode[0].flag; i++) {
        s = mainmode[i + 1].item;
        x = StrToInt(s);
        for (; *s && (*s != 'x'); s++);
        s += *s ? 1 : 0;
        y = StrToInt(s);
        if ((w == x) && (h == y)) {
          /* drop disabled flag */
          mainmode[1].flag ^= ITEM_HILITE;
          /* set new flag */
          mainmode[i + 1].flag |= ITEM_HILITE;
          /* found */
          infolist[SCR_IDX].stat = 1;
          break;
        }
      }
    }
    close(fl);
  } else {
    masksize = 0;
  }
}

void SavePatches(int n) {
struct ftime ft;
int fl, i, j;
WORD w;
BYTE b;
char *s;
  fl = open(gamelist[n].name, O_RDWR | O_BINARY);
  if (fl != -1) {
    getftime(fl, &ft);
    /* screen resolutions hack */
    infolist[SCR_IDX].stat = (mainmode[1].flag & ITEM_HILITE) ? 0 : 1;
    if (infolist[SCR_IDX].stat) {
      infolist[SCR_IDX].stat = 0;
      for (i = 1; i < mainmode[0].flag; i++) {
        if (mainmode[i + 1].flag & ITEM_HILITE) {
          s = mainmode[i + 1].item;
          w = StrToInt(s);
          datalist[infolist[SCR_IDX].idx + 3].data[1] = (w & 0xFF);
          datalist[infolist[SCR_IDX].idx + 4].data[1] = ((w >> 8) & 0xFF);
          /* create mask file */
          ReadBuildMask(w, 1);
          for (; *s && (*s != 'x'); s++);
          s += *s ? 1 : 0;
          w = StrToInt(s);
          datalist[infolist[SCR_IDX].idx + 5].data[1] = (w & 0xFF);
          datalist[infolist[SCR_IDX].idx + 6].data[1] = ((w >> 8) & 0xFF);
          /* found */
          infolist[SCR_IDX].stat = 1;
          break;
        }
      }
    } else {
      /* delete mask file */
      ReadBuildMask(0, 2);
    }
    for (i = 0; i < INFOSCOUNT; i++) {
      if (i < mainctrl[0].flag) {
        infolist[i].stat = (mainctrl[i + 1].flag & ITEM_HILITE) ? 1 : 0;
      }
      for (j = 0; j < infolist[i].count; j++) {
        /* offset must exists */
        if (datalist[infolist[i].idx + j].offs[n]) {
          /* relocs fix for PowerSlave */
          if (datalist[infolist[i].idx + j].offs[n] & 0x80000000) {
            datalist[infolist[i].idx + j].offs[n] ^= 0x80000000;
            b = 44;
          } else {
            b = 0;
          }
          b += datalist[infolist[i].idx + j].data[infolist[i].stat];
          lseek(fl, datalist[infolist[i].idx + j].offs[n], SEEK_SET);
          write(fl, &b, 1);
        }
      }
    }
    setftime(fl, &ft);
    close(fl);
  }
}

/*
  sprintf() adds a lot of code
  (not only the function itself but the error strings for floating types, etc.)
  so avoid to use that function
  arguments:
  buf - output buffer
  len - length in characters of output buffer (including last zero tail char)
  ... - strings to merge, last one MUST be NULL (used as arguments terminator)
*/
void StrMerge(char *buf, WORD len, ...) {
va_list args;
char *s;
  /* sanity checks */
  if (buf && len) {
    va_start(args, len);
    /* reserved 1 char for tail null */
    len--;
    /* merge all strings */
    do {
      s = va_arg(args, char *);
      if (s) {
        while (*s && len) {
          *buf = *s;
          buf++;
          s++;
          len--;
        }
      }
    } while (s && len);
    va_end(args);
    /* tail null */
    *buf = 0;
  }
}

int main(void) {
char s[34];
int i, n;
  /* startup */
  InitSetup(maininfo[1].item, maininfo[2].item);
  /* find required files */
  maingame[0].item = "Game selection";
  maingame[0].hint = "Available files";
  maingame[0].flag = 0;
  for (i = 0; i < GAMESCOUNT; i++) {
    if (FileExists(gamelist[i].name, gamelist[i].size)) {
      maingame[0].flag++;
      maingame[maingame[0].flag].item = gamelist[i].name;
      maingame[maingame[0].flag].hint = gamelist[i].game;
      maingame[maingame[0].flag].flag = 0;
    }
  }
  /* if no game executables found */
  i = maingame[0].flag;
  if (!i) {
    maingame[0].flag = 1 | MENU_NFOBOX;
    maingame[0].item = "Files not found";
    maingame[0].hint = NULL;
    maingame[1].item = "No supported game executables found!";
    maingame[1].hint = NULL;
  }
  /* do not show dialog for single executable */
  if (i != 1) {
    i = DrawMenu(maingame, NULL);
  }
  /* user actually selects something */
  if (i) {
    /* generate menu subheader */
    StrMerge(s, 34, "Current Filename ( ", maingame[i].item, " )", NULL);
    mainmenu[0].hint = s;
    /* get index of the selected item */
    n = i;
    for (i = 0; i < GAMESCOUNT; i++) {
      if (MatchItem(gamelist[i].name, maingame[n].item)) {
        n = i;
        break;
      }
    }
    /* load patches */
    LoadPatches(n);
    /* show main menu */
    i = DrawMenu(mainmenu, &MenuProc);
    /* if save changes or yes in additional menu */
    if ((i == 4) || ((i == 0) && (DrawMenu(mainexit, NULL) == 1))) {
      /* save patches */
      SavePatches(n);
    }
  }
  /* cleanup */
  FreeSetup();
  return(0);
}
